==> unicore-gateway6-4.3.0-2.src.tar.gz: src/main/java/eu/unicore/gateway/POSTHandler.java <==
/**
 * Copyright (c) 2005, Forschungszentrum Juelich
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met: 
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the Forschungszentrum Juelich nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


package eu.unicore.gateway;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URI;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.zip.GZIPInputStream;

import javanet.staxutils.XMLStreamEventWriter;

import javax.servlet.ServletException;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.stream.events.XMLEvent;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.ByteArrayRequestEntity;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.apache.log4j.Logger;
import org.codehaus.stax2.XMLOutputFactory2;
import org.codehaus.stax2.evt.XMLEventFactory2;

import com.ctc.wstx.stax.WstxEventFactory;

import eu.unicore.gateway.base.RawMessageExchange;
import eu.unicore.gateway.client.HttpClientFactory;
import eu.unicore.gateway.soap.SoapFault;
import eu.unicore.gateway.soap.SoapVersion;
import eu.unicore.gateway.soap.SoapFault.FaultCode;
import eu.unicore.gateway.util.AbstractStreamReaderRequestEntity;
import eu.unicore.gateway.util.BufferingProxyReader;
import eu.unicore.gateway.util.LogUtil;

/**
 * Main entry point for SOAP processing.
 * 
 * @author roger
 * @author golbi
 * 
 * TODO we could optimize: if the client accepts gzipped response and server provided it as 
 * gzipped (a common case we could forward as gzipped. However it would be troublesome as
 * jetty GZIP out handler should be disabled in this case and encoding set manually.
 * 
 */
public class POSTHandler
{
	private static final Logger log = LogUtil.getLogger(LogUtil.GATEWAY, POSTHandler.class);
	private final boolean chunked;
	private final SiteOrganiser organiser;
	private final IConsignorProducer consignorProducer;
	private final HttpClientFactory clientFactory;
	private final String serverUri;
	private HeadersParser hdrParser;
	private final static int FORWARD_BUF_SIZE = 10240;


	public POSTHandler(SiteOrganiser organiser, IConsignorProducer consignorProducer, 
		HttpClientFactory clientFactory, boolean chunked, String serverUri)
	{
		this.organiser = organiser;
		this.consignorProducer = consignorProducer;
		this.chunked = chunked;
		this.clientFactory = clientFactory;
		this.serverUri = serverUri;
		hdrParser = new HeadersParser(serverUri);
	}

	
	/**
	 * The message will be read and cached until either the headers have been fully read,
	 * or the destination can be figured out, from the WS-Addressing headers. 
	 * <p>	 
	 * Then, the cached XML will be written to the output, with the consignor assertion inserted
	 * just at the header beginning. 
	 * <p>
	 * If there is no header one is created (to carry the consignor assertion).
	 */
	public void invoke(final RawMessageExchange exchange) throws Exception
	{
		String destination = "?";
		try
		{
			hdrParser.parseHeaders(exchange);

			//get the certpath of the ssl connection (null if not ssl connection)
			final X509Certificate[] certs = (X509Certificate[])exchange.getProperty(
				RawMessageExchange.X509);
			
			VSite site=figureOutDestination(exchange);
			destination = exchange.getDestination();
			URI uri = URI.create(destination);

			//We stream the content through.
			//If non-chunked mode is turned on (for debugging) 
			//we must cache as we need to know the content-length
			RequestEntity requestentity = null;
			if (this.chunked)
			{
				requestentity = new AbstractStreamReaderRequestEntity()
				{
					@Override
					public void writeRequest(OutputStream os) throws IOException
					{
						try
						{
							writeToOutputStream(exchange, os, certs, 
								consignorProducer, log);
						}
						catch (XMLStreamException e)
						{
							IOException ioe = new IOException();
							ioe.initCause(e);
							throw ioe;
						}
					}
				};
			}
			else
			{
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				writeToOutputStream(exchange, baos, certs, consignorProducer, log);
				requestentity = new ByteArrayRequestEntity(baos.toByteArray());
			}

			log.debug("Dispatching to: " + destination);
			PostMethod post = prepareForwardedPOST(uri, exchange, requestentity);

			callServiceAndForwardResponse(post, uri, exchange, site);
			log.debug("Exchange with " + destination + " successfully completed");
		} catch (SoapFault e)
		{
			Throwable cause = e.getCause() == null ? e : e.getCause();
			LogUtil.logException("Problem during dispatch to " + destination,
				cause, log);
			if (exchange.getSoapVersion() != null)
				e.writeFaultToWriter(exchange.getWriter(), exchange.getSoapVersion());
			else
				throw new ServletException("The request doesn't look like a valid SOAP request.");
		}
	}


	
	private PostMethod prepareForwardedPOST(URI uri, RawMessageExchange exchange, RequestEntity reqEntity) 
		throws Exception
	{
		StringBuilder path=new StringBuilder(uri.getPath());
		if(uri.getQuery()!=null){
			path.append("?");
			path.append(uri.getQuery());
		}
		PostMethod post = clientFactory.makePostMethod(
			path.toString(), reqEntity);

		String contentType = (String)exchange.getProperty(RawMessageExchange.CONTENT_TYPE);
		if (contentType != null)
			post.setRequestHeader("Content-type", contentType);
		else
			post.setRequestHeader("Content-type", "text/xml; charset=UTF-8");
		
		String soapAction = (String)exchange.getProperty(RawMessageExchange.SOAP_ACTION);
		if (soapAction!=null)
		{
			post.setRequestHeader("SOAPAction", soapAction);
		} else
		{ 
			//If no SOAPAction in HTTP header let's try to get it from SOAP 
			//Header WSA element and put it as HTTP header so signature code on final server
			//will get it without parsing the request
			soapAction = exchange.getWsaAction();
			if (soapAction != null)
			{
				if(log.isDebugEnabled()) 
					log.debug("Generating HTTP SOAPAction from SOAP header: "+soapAction);
				post.setRequestHeader("SOAPAction", soapAction);
			}
		}
		String clientIP=(String)exchange.getProperty(RawMessageExchange.REMOTE_IP);
		if(clientIP!=null){
			if(log.isDebugEnabled()) 
				log.debug("Generating X-UNICORE-Consignor-IP header: "+clientIP);
			post.setRequestHeader("X-UNICORE-Consignor-IP", clientIP);
		}
		return post;
	}
	
	private void callServiceAndForwardResponse(PostMethod post, URI serviceUri, 
		RawMessageExchange exchange, VSite site) throws Exception
	{
		try
		{
			try
			{
				HttpClient client =null;
				synchronized(site){
					client = site.getClient();
					if(client==null){
						client = clientFactory.makeHttpClient(serviceUri.toURL());
						site.setClient(client);	
					}
				}
				//client.getParams().setParameter("http.protocol.expect-continue", true);
				client.executeMethod(post);
			} catch (Exception e)
			{
				String reason= "Problem when forwarding a client request to a VSite: "
					+ e.toString();
				log.debug("Got exception when forwarding a client request to the VSite " + 
					serviceUri + ": " + e);
				throw new SoapFault(serverUri, FaultCode.RECEIVER, reason, e);				
			}
			
			log.debug("Status of dispatch to service: " + post.getStatusLine());
			int status=post.getStatusCode();
			if(status>=400 && status<500){
				String reason= "Vsite returned error: "+post.getStatusLine();
				log.debug("Got error when forwarding a client request to the Vsite " + serviceUri + ": " + reason);
				throw new SoapFault(serverUri, FaultCode.RECEIVER, reason);			
			}
			
			try 
			{
				String contentEncoding = (post.getResponseHeader("Content-Encoding") != null) ? 
					post.getResponseHeader("Content-Encoding").getValue() : null;

				InputStream ris = "gzip".equalsIgnoreCase(contentEncoding) ? 
					new GZIPInputStream(post.getResponseBodyAsStream()) : 
					post.getResponseBodyAsStream();
				String responseCharSet = post.getResponseCharSet();
				forwardResponse(ris, responseCharSet, exchange, log);
			} catch (Exception e)
			{
				String reason = "Problem when forwarding a response " +
						"from a VSite to a client: " + e.toString();
				log.debug("Got exception when forwarding a response from the VSite " + 
					serviceUri + ": " + e);
				throw new SoapFault(serverUri, FaultCode.RECEIVER, reason, e);
			}
		} finally
		{
			post.releaseConnection();
		}
	}

	

	//places destination in exchange and returns the vsite
	private VSite figureOutDestination(RawMessageExchange ex) throws Exception
	{
		String address = ex.getWsaToAddress();
		String clientIP = (String)ex.getProperty(RawMessageExchange.REMOTE_IP);
		if (address == null)
		{
			address = ex.getRequestURL();
			log.debug("No wsa:To header found, using HTTP URL " + address);
		}
		
		VSite vsite = this.organiser.match(address, clientIP);
		if (vsite == null)
		{
			throw new SoapFault(serverUri, FaultCode.SENDER, "Cannot send message to " + address + 
				", such a site is not registered with this gateway.", null);
		}
		String destination=vsite.resolve(address);
		ex.setDestination(destination);
		return vsite;
	}



	/*
	 * The methods below are made static as must be are throughly tested.  
	 */
	public static void forwardResponse(InputStream ris, String responseCharSet,
		RawMessageExchange exchange, Logger log) throws Exception
	{
		Writer target = exchange.getWriter();
		byte[] buf = new byte[FORWARD_BUF_SIZE];
		int r = ris.read(buf, 0, buf.length);
		while (r > 0)
		{
			String s = new String(buf, 0, r, responseCharSet);
			if (log.isTraceEnabled())
				log.trace("Response part: " + s);
			target.write(s);
			r = ris.read(buf, 0, buf.length);
		}
		target.flush();
	}

	/**
	 * Copies data read from exchange to the os. At the marked position (in ex) consignor token is inserted. 
	 */
	public static void writeToOutputStream(RawMessageExchange ex, OutputStream os, X509Certificate[] certs, 
		IConsignorProducer producer, Logger log) 
		throws XMLStreamException, IOException
	{
		try
		{
			BufferingProxyReader reader = ex.getReader();

			//replay buffered data up to the marked position
			int boundary = reader.getMarkedPos();
			reader.replay(os, 0, boundary, "UTF-8");
			
			//add consignor token
			ByteArrayOutputStream baos = new ByteArrayOutputStream(4096);
			writeConsignorAssertionToBaos(baos, !ex.isHeaderPresent(), certs, 
				ex.getSoapVersion(), producer, log);
			baos.writeTo(os);
			
			//replay the rest of the buffer
			reader.replayRest(os, boundary, "UTF-8");
			
			//copy the rest (more or less the SOAP body)
			char[] buf = new char[1024];
			Writer w = new OutputStreamWriter(os);
			int r = reader.read(buf);
			while (r > 0)
			{
				w.write(buf, 0, r);
				r = reader.read(buf);
			}
			w.close();
		}
		catch (XMLStreamException e){
			LogUtil.logException("Error writing to destination.", e, log);
			throw e;
		}
	}
	
	private static void writeConsignorAssertionToBaos(ByteArrayOutputStream baos, boolean createHeader, 
		X509Certificate consignorCC[], SoapVersion soapVersion, IConsignorProducer consignorProducer, Logger log) throws XMLStreamException
	{
		XMLOutputFactory xof = XMLOutputFactory2.newInstance();
		xof.setProperty(XMLOutputFactory2.IS_REPAIRING_NAMESPACES, false);
		XMLStreamWriter xsw = xof.createXMLStreamWriter(baos);
		XMLEventWriter xew = new XMLStreamEventWriter(xsw);
		writeConsignorAssertion(xew, createHeader, consignorCC, soapVersion, consignorProducer, log);
		xsw.close();
	}

	private static void writeConsignorAssertion(XMLEventWriter xew, boolean createHeader, 
		X509Certificate consignorCC[], SoapVersion soapVersion, 
		IConsignorProducer consignorProducer, Logger log)
	{
		try{
			XMLEventFactory2 events = null;
			List<XMLEvent> events2 = consignorProducer.getConsignorAssertion(consignorCC,soapVersion);
			if (createHeader)
			{
				log.debug("Creating header");
				events=new WstxEventFactory();
				QName header = soapVersion.getHeader();
				xew.add(events.createStartElement(header.getPrefix(),header.getNamespaceURI(),header.getLocalPart()));
			}
			for (XMLEvent event: events2)
			{
				xew.add(event);
			}
			if (createHeader)
			{
				QName header = soapVersion.getHeader();
				xew.add(events.createEndElement(header.getPrefix(),header.getNamespaceURI(),header.getLocalPart()));			
			}
		} catch (Exception e){
			LogUtil.logException("Can't produce Consignor assertion. Shouldn't happen!", e, log);
		}
	}
}

==> unicore-gateway6-5.0.0-0.src.tar.gz: src/main/java/eu/unicore/gateway/POSTHandler.java <==
/**
 * Copyright (c) 2005, Forschungszentrum Juelich
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, are
 * permitted provided that the following conditions are met: 
 * 
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *  * Neither the name of the Forschungszentrum Juelich nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


package eu.unicore.gateway;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.URI;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.zip.GZIPInputStream;

import javanet.staxutils.XMLStreamEventWriter;

import javax.servlet.ServletException;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLEventWriter;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;
import javax.xml.stream.events.XMLEvent;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.methods.ByteArrayRequestEntity;
import org.apache.commons.httpclient.methods.PostMethod;
import org.apache.commons.httpclient.methods.RequestEntity;
import org.apache.log4j.Logger;
import org.codehaus.stax2.XMLOutputFactory2;
import org.codehaus.stax2.evt.XMLEventFactory2;

import com.ctc.wstx.stax.WstxEventFactory;

import eu.unicore.gateway.base.RawMessageExchange;
import eu.unicore.gateway.client.HttpClientFactory;
import eu.unicore.gateway.soap.SoapFault;
import eu.unicore.gateway.soap.SoapVersion;
import eu.unicore.gateway.soap.SoapFault.FaultCode;
import eu.unicore.gateway.util.AbstractStreamReaderRequestEntity;
import eu.unicore.gateway.util.BufferingProxyReader;
import eu.unicore.gateway.util.LogUtil;

/**
 * Main entry point for SOAP processing.
 * 
 * @author roger
 * @author golbi
 * 
 * TODO we could optimize: if the client accepts gzipped response and server provided it as 
 * gzipped (a common case we could forward as gzipped. However it would be troublesome as
 * jetty GZIP out handler should be disabled in this case and encoding set manually.
 * 
 */
public class POSTHandler
{
	private static final Logger log = LogUtil.getLogger(LogUtil.GATEWAY, POSTHandler.class);
	private final boolean chunked;
	private final SiteOrganiser organiser;
	private final IConsignorProducer consignorProducer;
	private final HttpClientFactory clientFactory;
	private final String serverUri;
	private HeadersParser hdrParser;
	private final static int FORWARD_BUF_SIZE = 10240;


	public POSTHandler(SiteOrganiser organiser, IConsignorProducer consignorProducer, 
		HttpClientFactory clientFactory, boolean chunked, String serverUri)
	{
		this.organiser = organiser;
		this.consignorProducer = consignorProducer;
		this.chunked = chunked;
		this.clientFactory = clientFactory;
		this.serverUri = serverUri;
		hdrParser = new HeadersParser(serverUri);
	}

	
	/**
	 * The message will be read and cached until either the headers have been fully read,
	 * or the destination can be figured out, from the WS-Addressing headers. 
	 * <p>	 
	 * Then, the cached XML will be written to the output, with the consignor assertion inserted
	 * just at the header beginning. 
	 * <p>
	 * If there is no header one is created (to carry the consignor assertion).
	 */
	public void invoke(final RawMessageExchange exchange) throws Exception
	{
		String destination = "?";
		try
		{
			hdrParser.parseHeaders(exchange);

			//get the certpath of the ssl connection (null if not ssl connection)
			final X509Certificate[] certs = (X509Certificate[])exchange.getProperty(
				RawMessageExchange.X509);
			final String clientIP=(String)exchange.getProperty(RawMessageExchange.REMOTE_IP);
			
			VSite site=figureOutDestination(exchange);
			destination = exchange.getDestination();
			URI uri = URI.create(destination);

			//We stream the content through.
			//If non-chunked mode is turned on (for debugging) 
			//we must cache as we need to know the content-length
			RequestEntity requestentity = null;
			if (this.chunked)
			{
				requestentity = new AbstractStreamReaderRequestEntity()
				{
					@Override
					public void writeRequest(OutputStream os) throws IOException
					{
						try
						{
							writeToOutputStream(exchange, os, certs, clientIP, 
								consignorProducer, log);
						}
						catch (XMLStreamException e)
						{
							IOException ioe = new IOException();
							ioe.initCause(e);
							throw ioe;
						}
					}
				};
			}
			else
			{
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				writeToOutputStream(exchange, baos, certs, clientIP, consignorProducer, log);
				requestentity = new ByteArrayRequestEntity(baos.toByteArray());
			}

			log.debug("Dispatching to: " + destination);
			PostMethod post = prepareForwardedPOST(uri, exchange, requestentity);

			callServiceAndForwardResponse(post, uri, exchange, site);
			log.debug("Exchange with " + destination + " successfully completed");
		} catch (SoapFault e)
		{
			Throwable cause = e.getCause() == null ? e : e.getCause();
			LogUtil.logException("Problem during dispatch to " + destination,
				cause, log);
			if (exchange.getSoapVersion() != null)
				e.writeFaultToWriter(exchange.getWriter(), exchange.getSoapVersion());
			else
				throw new ServletException("The request doesn't look like a valid SOAP request.");
		}
	}


	
	private PostMethod prepareForwardedPOST(URI uri, RawMessageExchange exchange, RequestEntity reqEntity) 
		throws Exception
	{
		StringBuilder path=new StringBuilder(uri.getPath());
		if(uri.getQuery()!=null){
			path.append("?");
			path.append(uri.getQuery());
		}
		PostMethod post = clientFactory.makePostMethod(
			path.toString(), reqEntity);

		String contentType = (String)exchange.getProperty(RawMessageExchange.CONTENT_TYPE);
		if (contentType != null)
			post.setRequestHeader("Content-type", contentType);
		else
			post.setRequestHeader("Content-type", "text/xml; charset=UTF-8");
		
		String soapAction = (String)exchange.getProperty(RawMessageExchange.SOAP_ACTION);
		if (soapAction!=null)
		{
			post.setRequestHeader("SOAPAction", soapAction);
		} else
		{ 
			//If no SOAPAction in HTTP header let's try to get it from SOAP 
			//Header WSA element and put it as HTTP header so signature code on final server
			//will get it without parsing the request
			soapAction = exchange.getWsaAction();
			if (soapAction != null)
			{
				if(log.isDebugEnabled()) 
					log.debug("Generating HTTP SOAPAction from SOAP header: "+soapAction);
				post.setRequestHeader("SOAPAction", soapAction);
			}
		}
		return post;
	}
	
	private void callServiceAndForwardResponse(PostMethod post, URI serviceUri, 
		RawMessageExchange exchange, VSite site) throws Exception
	{
		try
		{
			try
			{
				HttpClient client =null;
				synchronized(site){
					client = site.getClient();
					if(client==null){
						client = clientFactory.makeHttpClient(serviceUri.toURL());
						site.setClient(client);	
					}
				}
				//client.getParams().setParameter("http.protocol.expect-continue", true);
				client.executeMethod(post);
			} catch (Exception e)
			{
				String reason= "Problem when forwarding a client request to a VSite: "
					+ e.toString();
				log.debug("Got exception when forwarding a client request to the VSite " + 
					serviceUri + ": " + e);
				throw new SoapFault(serverUri, FaultCode.RECEIVER, reason, e);				
			}
			
			log.debug("Status of dispatch to service: " + post.getStatusLine());
			int status=post.getStatusCode();
			if(status>=400 && status<500){
				String reason= "Vsite returned error: "+post.getStatusLine();
				log.debug("Got error when forwarding a client request to the Vsite " + serviceUri + ": " + reason);
				throw new SoapFault(serverUri, FaultCode.RECEIVER, reason);			
			}
			
			try 
			{
				String contentEncoding = (post.getResponseHeader("Content-Encoding") != null) ? 
					post.getResponseHeader("Content-Encoding").getValue() : null;

				InputStream ris = "gzip".equalsIgnoreCase(contentEncoding) ? 
					new GZIPInputStream(post.getResponseBodyAsStream()) : 
					post.getResponseBodyAsStream();
				String responseCharSet = post.getResponseCharSet();
				forwardResponse(ris, responseCharSet, exchange, log);
			} catch (Exception e)
			{
				String reason = "Problem when forwarding a response " +
						"from a VSite to a client: " + e.toString();
				log.debug("Got exception when forwarding a response from the VSite " + 
					serviceUri + ": " + e);
				throw new SoapFault(serverUri, FaultCode.RECEIVER, reason, e);
			}
		} finally
		{
			post.releaseConnection();
		}
	}

	

	//places destination in exchange and returns the vsite
	private VSite figureOutDestination(RawMessageExchange ex) throws Exception
	{
		String address = ex.getWsaToAddress();
		String clientIP = (String)ex.getProperty(RawMessageExchange.REMOTE_IP);
		if (address == null)
		{
			address = ex.getRequestURL();
			log.debug("No wsa:To header found, using HTTP URL " + address);
		}
		
		VSite vsite = this.organiser.match(address, clientIP);
		if (vsite == null)
		{
			throw new SoapFault(serverUri, FaultCode.SENDER, "Cannot send message to " + address + 
				", such a site is not registered with this gateway.", null);
		}
		String destination=vsite.resolve(address);
		ex.setDestination(destination);
		return vsite;
	}



	/*
	 * The methods below are made static as must be are throughly tested.  
	 */
	public static void forwardResponse(InputStream ris, String responseCharSet,
		RawMessageExchange exchange, Logger log) throws Exception
	{
		Writer target = exchange.getWriter();
		byte[] buf = new byte[FORWARD_BUF_SIZE];
		int r = ris.read(buf, 0, buf.length);
		while (r > 0)
		{
			String s = new String(buf, 0, r, responseCharSet);
			if (log.isTraceEnabled())
				log.trace("Response part: " + s);
			target.write(s);
			r = ris.read(buf, 0, buf.length);
		}
		target.flush();
	}

	/**
	 * Copies data read from exchange to the os. At the marked position (in ex) consignor token is inserted. 
	 */
	public static void writeToOutputStream(RawMessageExchange ex, OutputStream os, X509Certificate[] certs, 
			String ip, IConsignorProducer producer, Logger log) throws XMLStreamException, IOException
	{
		try
		{
			BufferingProxyReader reader = ex.getReader();

			//replay buffered data up to the marked position
			int boundary = reader.getMarkedPos();
			reader.replay(os, 0, boundary, "UTF-8");
			
			//add consignor token
			ByteArrayOutputStream baos = new ByteArrayOutputStream(4096);
			writeConsignorAssertionToBaos(baos, !ex.isHeaderPresent(), certs, ip, 
				ex.getSoapVersion(), producer, log);
			baos.writeTo(os);
			
			//replay the rest of the buffer
			reader.replayRest(os, boundary, "UTF-8");
			
			//copy the rest (more or less the SOAP body)
			char[] buf = new char[1024];
			Writer w = new OutputStreamWriter(os);
			int r = reader.read(buf);
			while (r > 0)
			{
				w.write(buf, 0, r);
				r = reader.read(buf);
			}
			w.close();
		}
		catch (XMLStreamException e){
			LogUtil.logException("Error writing to destination.", e, log);
			throw e;
		}
	}
	
	private static void writeConsignorAssertionToBaos(ByteArrayOutputStream baos, boolean createHeader, 
		X509Certificate consignorCC[], String ip, SoapVersion soapVersion, 
		IConsignorProducer consignorProducer, Logger log) throws XMLStreamException
	{
		XMLOutputFactory xof = XMLOutputFactory2.newInstance();
		xof.setProperty(XMLOutputFactory2.IS_REPAIRING_NAMESPACES, false);
		XMLStreamWriter xsw = xof.createXMLStreamWriter(baos);
		XMLEventWriter xew = new XMLStreamEventWriter(xsw);
		writeConsignorAssertion(xew, createHeader, consignorCC, ip, soapVersion, consignorProducer, log);
		xsw.close();
	}

	private static void writeConsignorAssertion(XMLEventWriter xew, boolean createHeader, 
		X509Certificate consignorCC[], String ip, SoapVersion soapVersion, 
		IConsignorProducer consignorProducer, Logger log)
	{
		try{
			XMLEventFactory2 events = null;
			List<XMLEvent> events2 = consignorProducer.getConsignorAssertion(consignorCC, ip, soapVersion);
			if (createHeader)
			{
				log.debug("Creating header");
				events=new WstxEventFactory();
				QName header = soapVersion.getHeader();
				xew.add(events.createStartElement(header.getPrefix(),header.getNamespaceURI(),header.getLocalPart()));
			}
			for (XMLEvent event: events2)
			{
				xew.add(event);
			}
			if (createHeader)
			{
				QName header = soapVersion.getHeader();
				xew.add(events.createEndElement(header.getPrefix(),header.getNamespaceURI(),header.getLocalPart()));			
			}
		} catch (Exception e){
			LogUtil.logException("Can't produce Consignor assertion. Shouldn't happen!", e, log);
		}
	}
}
